@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,14 +24,15 @@
 # Invariants: The periodic callback assumes pcm_ok. If we don't have
 # at least pcm_ok, we have no sound whatsoever.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+from future.utils import raise_
 
-import renpy.audio  # @UnusedImport
-import renpy.display  # @UnusedImport
-from renpy import six
+import renpy.audio # @UnusedImport
+import renpy.display # @UnusedImport
 
 import time
-import pygame_sdl2  # @UnusedImport
+import pygame_sdl2 # @UnusedImport
 import os
 import re
 import threading
@@ -40,13 +41,7 @@ import io
 
 # Import the appropriate modules, or set them to None if we cannot.
 
-disable = os.environ.get("RENPY_DISABLE_SOUND", "")
-
-if not disable:
-    import renpy.audio.renpysound as renpysound
-else:
-    renpysound = None
-
+import renpy.audio.renpysound as renpysound
 
 # This is True if we were able to sucessfully enable the pcm audio.
 pcm_ok = None
@@ -70,11 +65,20 @@ def load(fn):
     Returns a file-like object for the given filename.
     """
 
-    rv = renpy.loader.load(fn)
+    try:
+        rv = renpy.loader.load(fn)
+    except renpy.webloader.DownloadNeeded as exception:
+        if exception.rtype == 'music':
+            renpy.webloader.enqueue(exception.relpath, 'music', None)
+        elif exception.rtype == 'voice':
+            # prediction failed, too late
+            pass
+        # temporary 1s placeholder, will retry loading when looping:
+        rv = open(os.path.join(renpy.config.commondir, '_dl_silence.ogg'), 'rb')
     return rv
 
 
-class AudioData(unicode):
+class AudioData(str):
     """
     :doc: audio
 
@@ -99,7 +103,7 @@ class AudioData(unicode):
     """
 
     def __new__(cls, data, filename):
-        rv = unicode.__new__(cls, filename)
+        rv = str.__new__(cls, filename)
         rv.data = data
         return rv
 
@@ -107,7 +111,7 @@ class AudioData(unicode):
         pass
 
     def __reduce__(self):
-        return(AudioData, (self.data, unicode(self)))
+        return(AudioData, (self.data, str(self)))
 
 
 class QueueEntry(object):
@@ -115,11 +119,12 @@ class QueueEntry(object):
     A queue entry object.
     """
 
-    def __init__(self, filename, fadein, tight, loop):
+    def __init__(self, filename, fadein, tight, loop, relative_volume):
         self.filename = filename
         self.fadein = fadein
         self.tight = tight
         self.loop = loop
+        self.relative_volume = relative_volume
 
 
 class MusicContext(renpy.python.RevertableObject):
@@ -132,6 +137,7 @@ class MusicContext(renpy.python.RevertableObject):
     __version__ = 0
 
     pause = False
+    tertiary_volume = 1.0
 
     def __init__(self):
 
@@ -149,6 +155,9 @@ class MusicContext(renpy.python.RevertableObject):
         # The secondary volume.
         self.secondary_volume = 1.0
 
+        # The tertiary volume.
+        self.tertiary_volume = 1.0
+
         # The time the channel was ordered last changed.
         self.last_changed = 0
 
@@ -313,6 +322,18 @@ class Channel(object):
 
     context = property(get_context)
 
+    def copy_context(self):
+        """
+        Copies the MusicContext associated with this channel, updates the
+        ExecutionContext to point to the copy, and returns the copy.
+        """
+
+        mcd = renpy.game.context().music
+
+        ctx = self.get_context().copy()
+        mcd[self.name] = ctx
+        return ctx
+
     def split_filename(self, filename, looped):
         """
         Splits a filename into a filename, start time, and end time.
@@ -332,12 +353,27 @@ class Channel(object):
             except:
                 raise exception("expected float, got {!r}.".format(v))
 
+        def expect_channel():
+            if not spec:
+                raise exception("expected channel at end.")
+
+            v = spec.pop(0)
+
+            try:
+                return renpy.audio.audio.get_channel(v)
+            except:
+                raise exception("expected channel, got {!r}.".format(v))
+
+        if isinstance(filename, AudioData):
+            return filename, 0, -1
+
         m = re.match(r'<(.*)>(.*)', filename)
         if not m:
-            return filename, 0, -1
+            return self.file_prefix + filename + self.file_suffix, 0, -1
 
         spec = m.group(1)
         fn = m.group(2)
+        fn = self.file_prefix + fn + self.file_suffix
 
         spec = spec.split()
 
@@ -352,6 +388,13 @@ class Channel(object):
                 start = expect_float()
             elif clause == "to":
                 end = expect_float()
+            elif clause == "sync":
+                sync_channel = expect_channel()
+                t = sync_channel.get_pos()
+                if not t or t < 0:
+                    pass
+                else:
+                    start = t
             elif clause == "loop":
                 loop = expect_float()
             elif clause == "silence":
@@ -373,7 +416,14 @@ class Channel(object):
         """
 
         # Update the channel volume.
-        vol = self.chan_volume * renpy.game.preferences.volumes.get(self.mixer, 1.0)
+
+        mixer_volume = renpy.game.preferences.volumes.get(self.mixer, 1.0)
+
+        if renpy.game.preferences.self_voicing:
+            if self.mixer not in renpy.config.voice_mixers:
+                mixer_volume = mixer_volume * renpy.game.preferences.self_voicing_volume_drop
+
+        vol = self.chan_volume * mixer_volume
 
         if vol != self.actual_volume:
             renpysound.set_volume(self.number, vol)
@@ -448,13 +498,18 @@ class Channel(object):
             try:
                 filename, start, end = self.split_filename(topq.filename, topq.loop)
 
+                if renpy.config.audio_filename_callback is not None:
+                    filename = renpy.config.audio_filename_callback(filename)
+
+                self.set_tertiary_volume(topq.relative_volume)
+
                 if (end >= 0) and ((end - start) <= 0) and self.queue:
                     continue
 
                 if isinstance(topq.filename, AudioData):
                     topf = io.BytesIO(topq.filename.data)
                 else:
-                    topf = load(self.file_prefix + filename + self.file_suffix)
+                    topf = load(filename)
 
                 renpysound.set_video(self.number, self.movie)
 
@@ -485,14 +540,14 @@ class Channel(object):
             if self.loop:
                 for i in self.loop:
                     if topq is not None:
-                        newq = QueueEntry(i, 0, topq.tight, True)
+                        newq = QueueEntry(i, 0, topq.tight, True, topq.relative_volume)
                     else:
-                        newq = QueueEntry(i, 0, False, True)
+                        newq = QueueEntry(i, 0, False, True, 1.0)
 
                     self.queue.append(newq)
             # Try callback:
             elif self.callback:
-                self.callback()  # E1102
+                self.callback() # E1102
 
         want_pause = self.context.pause or global_pause
 
@@ -543,12 +598,13 @@ class Channel(object):
 
             if self.secondary_volume_time != self.context.secondary_volume_time:
                 self.secondary_volume_time = self.context.secondary_volume_time
+                result_volume = self.context.secondary_volume * self.context.tertiary_volume
                 renpysound.set_secondary_volume(self.number,
-                                                self.context.secondary_volume,
+                                                result_volume,
                                                 0)
 
         if not self.queue and self.callback:
-            self.callback()  # E1102
+            self.callback() # E1102
 
     def fadeout(self, secs):
         """
@@ -569,16 +625,23 @@ class Channel(object):
             else:
                 renpysound.fadeout(self.number, secs)
 
-    def enqueue(self, filenames, loop=True, synchro_start=False, fadein=0, tight=None, loop_only=False):
+    def reload(self):
+        """
+        Causes this channel to be stopped in a way that looped audio will be
+        reloaded and restarted.
+        """
+
+        with lock:
+            renpysound.dequeue(self.number, True)
+            renpysound.stop(self.number)
+
+    def enqueue(self, filenames, loop=True, synchro_start=False, fadein=0, tight=None, loop_only=False, relative_volume=1.0):
 
         with lock:
 
             for filename in filenames:
                 filename, _, _ = self.split_filename(filename, False)
-                renpy.game.persistent._seen_audio[filename] = True  # @UndefinedVariable
-
-            if not pcm_ok:
-                return
+                renpy.game.persistent._seen_audio[filename] = True # @UndefinedVariable
 
             if not loop_only:
 
@@ -588,7 +651,7 @@ class Channel(object):
                 self.keep_queue += 1
 
                 for filename in filenames:
-                    qe = QueueEntry(filename, int(fadein * 1000), tight, False)
+                    qe = QueueEntry(filename, fadein, tight, False, relative_volume)
                     self.queue.append(qe)
 
                     # Only fade the first thing in.
@@ -660,7 +723,12 @@ class Channel(object):
 
             if pcm_ok:
                 self.secondary_volume_time = self.context.secondary_volume_time
-                renpysound.set_secondary_volume(self.number, self.context.secondary_volume, delay)
+                result_volume = self.context.secondary_volume * self.context.tertiary_volume
+                renpysound.set_secondary_volume(self.number, result_volume, delay)
+
+    def set_tertiary_volume(self, volume):
+        self.context.tertiary_volume = volume
+        self.set_secondary_volume(self.context.secondary_volume, 0)
 
     def pause(self):
         with lock:
@@ -683,10 +751,10 @@ class Channel(object):
 
         return renpysound.video_ready(self.number)
 
-
 # Use unconditional imports so these files get compiled during the build
 # process.
 
+
 try:
     from renpy.audio.androidhw import AndroidVideoChannel
 except:
@@ -841,17 +909,21 @@ def init():
     global pcm_ok
     global mix_ok
 
-    if not renpy.config.sound:
+    if not renpy.config.sound or ("RENPY_DISABLE_SOUND" in os.environ):
         pcm_ok = False
         mix_ok = False
         return
 
+    if renpy.emscripten and renpy.config.webaudio:
+        import renpy.audio.webaudio as webaudio
+        renpysound.__dict__.update(webaudio.__dict__)
+
     if pcm_ok is None and renpysound:
         bufsize = 2048
         if renpy.emscripten:
             # Large buffer (and latency) as compromise to avoid sound jittering
-            bufsize = 8192  # works for me
-            #bufsize = 16384  # jitter/silence right after starting a sound
+            bufsize = 8192 # works for me
+            # bufsize = 16384  # jitter/silence right after starting a sound
 
         if 'RENPY_SOUND_BUFSIZE' in os.environ:
             bufsize = int(os.environ['RENPY_SOUND_BUFSIZE'])
@@ -860,9 +932,17 @@ def init():
             renpysound.init(renpy.config.sound_sample_rate, 2, bufsize, False, renpy.config.equal_mono)
             pcm_ok = True
         except:
+
             if renpy.config.debug_sound:
                 raise
-            pcm_ok = False
+
+            os.environ["SDL_AUDIODRIVER"] = "dummy"
+
+            try:
+                renpysound.init(renpy.config.sound_sample_rate, 2, bufsize, False, renpy.config.equal_mono)
+                pcm_ok = True
+            except:
+                pcm_ok = False
 
     # Find all of the mixers in the game.
     mixers = [ ]
@@ -886,7 +966,7 @@ def init():
         periodic_thread.start()
 
 
-def quit():  # @ReservedAssignment
+def quit(): # @ReservedAssignment
 
     global periodic_thread
     global periodic_thread_quit
@@ -922,7 +1002,6 @@ def quit():  # @ReservedAssignment
     mix_ok = None
 
 
-
 # The last-set pcm volume.
 pcm_volume = None
 
@@ -1006,7 +1085,6 @@ def periodic_pass():
 # The exception that's been thrown by the periodic thread.
 periodic_exc = None
 
-
 # Should we run the periodic thread now?
 run_periodic = False
 
@@ -1057,7 +1135,7 @@ def periodic():
             exc = periodic_exc
             periodic_exc = None
 
-            six.reraise(exc[0], exc[1], exc[2])
+            raise_(exc[0], exc[1], exc[2])
 
         run_periodic = True
         periodic_condition.notify()
@@ -1093,10 +1171,10 @@ def interact():
 
                 if c.loop:
                     if not filenames or c.get_playing() not in filenames:
-                        c.fadeout(renpy.config.fade_music)
+                        c.fadeout(renpy.config.context_fadeout_music or renpy.config.fade_music)
 
                 if filenames:
-                    c.enqueue(filenames, loop=True, synchro_start=True, tight=tight)
+                    c.enqueue(filenames, loop=True, synchro_start=False, tight=tight, fadein=renpy.config.context_fadein_music)
 
                 c.last_changed = ctx.last_changed
 
@@ -1119,6 +1197,18 @@ def rollback():
                 c.fadeout(0)
 
 
+def autoreload(_fn):
+    """
+    After a sound file has been changed, stop all sound (and let Ren'Py restart
+    the channels, as needed.)
+    """
+
+    for c in all_channels:
+        c.reload()
+
+    renpy.exports.restart_interaction()
+
+
 global_pause = False
 
 
