@@ -80,7 +80,7 @@ def compile_event(key, keydown):
     if part[0] == "joy" or part[0] == "pad":
         return "(False)"
 
-    MODIFIERS = { "keydown", "keyup", "repeat", "alt", "meta", "shift", "noshift", "ctrl" }
+    MODIFIERS = { "keydown", "keyup", "repeat", "alt", "meta", "shift", "noshift", "ctrl", "osctrl" }
     modifiers = set()
 
     while part[0] in MODIFIERS:
@@ -106,7 +106,7 @@ def compile_event(key, keydown):
 
     if key not in [ "K_LALT", "K_RALT" ]:
 
-        if "alt" in modifiers:
+        if "alt" in modifiers or (renpy.macintosh and "osctrl" in modifiers):
             rv += " and (ev.mod & %d)" % pygame.KMOD_ALT
         else:
             rv += " and not (ev.mod & %d)" % pygame.KMOD_ALT
@@ -120,7 +120,7 @@ def compile_event(key, keydown):
 
     if key not in [ "K_LCTRL", "K_RCTRL" ]:
 
-        if "ctrl" in modifiers:
+        if "ctrl" in modifiers or (not renpy.macintosh and "osctrl" in modifiers):
             rv += " and (ev.mod & %d)" % pygame.KMOD_CTRL
         else:
             rv += " and not (ev.mod & %d)" % pygame.KMOD_CTRL
@@ -673,7 +673,11 @@ class SayBehavior(renpy.display.layout.Null):
         if renpy.config.skipping and renpy.config.allow_skipping and renpy.store._skipping:
 
             if ev.type == renpy.display.core.TIMEEVENT and st >= skip_delay:
-                if renpy.game.preferences.skip_unseen:
+
+                if ev.modal:
+                    renpy.config.skipping = None
+                    renpy.exports.restart_interaction()
+                elif renpy.game.preferences.skip_unseen:
                     return True
                 elif renpy.config.skipping == "fast":
                     return True
@@ -1126,6 +1130,47 @@ def input_post_per_interact():
             i.caret_pos = len(content)
 
 
+class CaretBlink(renpy.display.core.Displayable):
+    """
+    A displayable that renders the caret.
+    """
+
+    def __init__(self, caret, caret_blink, **properties):
+
+        super(CaretBlink, self).__init__(**properties)
+        caret = renpy.easy.displayable(caret)
+
+        if caret._duplicatable:
+            caret = caret._duplicate()
+            caret._unique()
+
+        self.caret = caret
+        self.caret_blink = caret_blink
+
+        self.st_base = 0
+
+    def get_position(self):
+        return self.caret.get_position()
+
+    def visit(self):
+        return [ self.caret ]
+
+    def render(self, width, height, st, at):
+        st -= self.st_base
+
+        cr = renpy.display.render.render(self.caret, width, height, st, at)
+        rv = renpy.display.render.Render(0, height)
+
+        ttl = self.caret_blink - st % self.caret_blink
+
+        if ttl > self.caret_blink / 2.0:
+            rv.blit(cr, (0, 0))
+
+        renpy.display.render.redraw(self, ttl % (self.caret_blink / 2.))
+
+        return rv
+
+
 class Input(renpy.text.text.Text): # @UndefinedVariable
     """
     This is a Displayable that takes text as input.
@@ -1142,6 +1187,8 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
     value = None
     shown = False
 
+    st = 0
+
     def __init__(self,
                  default="",
                  length=None,
@@ -1157,10 +1204,14 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
                  pixel_width=None,
                  value=None,
                  copypaste=False,
+                 caret_blink=None,
                  **properties):
 
         super(Input, self).__init__("", style=style, replaces=replaces, substitute=False, **properties)
 
+        if caret_blink is None:
+            caret_blink = renpy.config.input_caret_blink
+
         if value:
             self.value = value
             changed = value.set_text
@@ -1188,7 +1239,12 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             if i.endswith("color"):
                 caretprops[i] = properties[i]
 
-        self.caret = renpy.display.image.Solid(xmaximum=1, style=style, **caretprops)
+        caret = renpy.display.image.Solid(xsize=1, style=style, **caretprops)
+
+        if caret_blink:
+            caret = CaretBlink(caret, caret_blink)
+
+        self.caret = caret
         self.caret_pos = len(self.content)
         self.old_caret_pos = self.caret_pos
 
@@ -1212,12 +1268,14 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
         old_content = self.content
 
         if new_content != self.content or editable != self.editable or edit:
+
             renpy.display.render.redraw(self, 0)
 
         self.editable = editable
 
         # Choose the caret.
         caret = self.style.caret
+
         if caret is None:
             caret = self.caret
 
@@ -1248,7 +1306,7 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
         def set_content(content):
 
             if content == "":
-                content = u"\u200b"
+                content = u" "
 
             if editable:
                 l = len(content)
@@ -1257,6 +1315,10 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             else:
                 self.set_text([self.prefix, content.replace("{", "{{"), self.suffix ])
 
+            if isinstance(self.caret, CaretBlink):
+                self.caret.st_base = self.st
+                renpy.display.render.redraw(self.caret, 0)
+
         set_content(new_content)
 
         if check_size and self.pixel_width:
@@ -1311,12 +1373,20 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
 
     def event(self, ev, x, y, st):
 
+        self.st = st
         self.old_caret_pos = self.caret_pos
 
         if not self.editable:
             return None
 
-        if (ev.type == pygame.KEYDOWN) and (pygame.key.get_mods() & pygame.KMOD_LALT) and (not ev.unicode):
+        edit_controls = any([
+            map_event(ev, "input_jump_word_left"),
+            map_event(ev, "input_jump_word_right"),
+            map_event(ev, "input_delete_word"),
+            map_event(ev, "input_delete_full"),
+        ])
+
+        if (ev.type == pygame.KEYDOWN) and (pygame.key.get_mods() & pygame.KMOD_LALT) and (not ev.unicode) and not edit_controls:
             return None
 
         l = len(self.content)
@@ -1354,6 +1424,21 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_jump_word_left"):
+            if self.caret_pos > 0:
+                space_pos = 0
+                prev_end = 0
+                for item in re.finditer(r"\s+", self.content[:self.caret_pos]):
+                    start, end = item.span()
+                    if end != self.caret_pos:
+                        space_pos = end
+                    prev_end = end
+                self.caret_pos = space_pos
+                self.update_text(self.content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_right"):
             if self.caret_pos < l:
                 self.caret_pos += 1
@@ -1362,6 +1447,19 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_jump_word_right"):
+            if self.caret_pos < l:
+                space_pos = l
+                for item in re.finditer(r"\s+", self.content[self.caret_pos + 1:]):
+                    start, end = item.span()
+                    space_pos = end
+                    break
+                self.caret_pos = min(space_pos + self.caret_pos + 1, l)
+                self.update_text(self.content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_delete"):
             if self.caret_pos < l:
                 content = self.content[0:self.caret_pos] + self.content[self.caret_pos + 1:l]
@@ -1370,6 +1468,29 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_delete_word"):
+            if self.caret_pos <= l:
+                space_pos = 0
+                for item in re.finditer(r"\s+", self.content[:self.caret_pos]):
+                    start, end = item.span()
+                    if end != self.caret_pos:
+                        space_pos = end
+                content = self.content[0:space_pos] + self.content[self.caret_pos:l]
+                self.caret_pos = space_pos
+                self.update_text(content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
+        elif map_event(ev, "input_delete_full"):
+            if self.caret_pos <= l:
+                content = self.content[self.caret_pos:l]
+                self.caret_pos = 0
+                self.update_text(content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_home"):
             self.caret_pos = 0
             self.update_text(self.content, self.editable)
@@ -1464,6 +1585,8 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             raise renpy.display.core.IgnoreEvent()
 
     def render(self, width, height, st, at):
+        self.st = st
+
         rv = super(Input, self).render(width, height, st, at)
 
         if self.editable:
@@ -1538,6 +1661,9 @@ class Adjustment(renpy.object.Object):
             This function is called with the adjustment object when
             the range of the adjustment is set by a viewport.
 
+            This function may be called multiple times, as part of the layout
+            process.
+
         `force_step`
             If True and this adjustment changes by dragging associated
             viewport or a bar, value will be changed only if the drag
@@ -1779,11 +1905,12 @@ class Bar(renpy.display.core.Displayable):
             if redraw is not None:
                 renpy.display.render.redraw(self, redraw)
 
-        xminimum = self.style.xminimum
-        yminimum = self.style.yminimum
+        xminimum, yminimum = renpy.display.layout.xyminimums(self.style, width, height)
 
         if xminimum is not None:
             width = max(width, xminimum)
+
+        if yminimum is not None:
             height = max(height, yminimum)
 
         # Store the width and height for the event function to use.
@@ -2190,6 +2317,9 @@ class MouseArea(renpy.display.core.Displayable):
         if renpy.display.focus.pending_focus_type == 'keyboard':
             is_hovered = False
 
+        if (ev.type == renpy.display.core.TIMEEVENT) and ev.modal:
+            is_hovered = False
+
         else:
             if self.style.focus_mask is not None:
                 crend = renpy.display.render.render(self.style.focus_mask, self.width, self.height, st, self.at_st_offset + st)
