@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,28 +19,36 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 # This file contains displayables that move, zoom, rotate, or otherwise
 # transform displayables. (As well as displayables that support them.)
 import math
-import types #@UnresolvedImport
+import types # @UnresolvedImport
 
-import renpy.display #@UnusedImport
+import renpy.display # @UnusedImport
 from renpy.display.layout import Container
 
-import renpy.display.accelerator
+from renpy.display.accelerator import transform_render
+from renpy.atl import position, any_object, bool_or_none, float_or_none, matrix, mesh
 
 # The null object that's used if we don't have a defined child.
 null = None
 
+
 def get_null():
     global null
 
     if null is None:
         null = renpy.display.layout.Null()
+        renpy.display.motion.null = null
 
     return null
 
 # Convert a position from cartesian to polar coordinates.
+
+
 def cartesian_to_polar(x, y, xaround, yaround):
     """
     Converts cartesian coordinates to polar coordinates.
@@ -57,6 +65,7 @@ def cartesian_to_polar(x, y, xaround, yaround):
 
     return angle, radius
 
+
 def polar_to_cartesian(angle, radius, xaround, yaround):
     """
     Converts polart coordinates to cartesian coordinates.
@@ -72,6 +81,7 @@ def polar_to_cartesian(angle, radius, xaround, yaround):
 
     return x, y
 
+
 def first_not_none(*args):
     """
     Returns the first argument that is not None.
@@ -85,70 +95,12 @@ def first_not_none(*args):
 
 class TransformState(renpy.object.Object):
 
-    nearest = None
-    xoffset = None
-    yoffset = None
-    inherited_xpos = None
-    inherited_ypos = None
-    inherited_xanchor = None
-    inherited_yanchor = None
-    transform_anchor = False
-    additive = 0.0
-    debug = None
-    events = True
-    crop_relative = False
-    xpan = None
-    ypan = None
-    xtile = 1
-    ytile = 1
+    last_angle = None
 
     def __init__(self):
-        self.alpha = 1
-        self.nearest = None
-        self.additive = 0.0
-        self.rotate = None
-        self.rotate_pad = True
-        self.transform_anchor = False
-        self.zoom = 1
-        self.xzoom = 1
-        self.yzoom = 1
-
-        self.xpos = None
-        self.ypos = None
-        self.xanchor = None
-        self.yanchor = None
-        self.xoffset = 0
-        self.yoffset = 0
-
-        self.xaround = 0.0
-        self.yaround = 0.0
-        self.xanchoraround = 0.0
-        self.yanchoraround = 0.0
-
-        self.xpan = None
-        self.ypan = None
-        self.xtile = 1
-        self.ytile = 1
-
-        self.subpixel = False
-
-        self.crop = None
-        self.crop_relative = False
-        self.corner1 = None
-        self.corner2 = None
-        self.size = None
-
-        self.delay = 0
-
-        self.debug = None
-        self.events = True
-
-        # Note: When adding a new property, we need to add it to:
-        # - take_state
-        # - diff
-        # - renpy.atl.PROPERTIES
-        # - Proxies in Transform
 
+        # Most fields on this object are set by add_property, at the bottom
+        # of this file.
 
         # An xpos (etc) inherited from our child overrides an xpos inherited
         # from an old transform, but not an xpos set in the current transform.
@@ -160,36 +112,24 @@ class TransformState(renpy.object.Object):
         self.inherited_xanchor = None
         self.inherited_yanchor = None
 
+        # The last angle that was rotated to.
+        self.last_angle = None
+
     def take_state(self, ts):
 
-        self.nearest = ts.nearest
-        self.alpha = ts.alpha
-        self.additive = ts.additive
-        self.rotate = ts.rotate
-        self.rotate_pad = ts.rotate_pad
-        self.transform_anchor = ts.transform_anchor
-        self.zoom = ts.zoom
-        self.xzoom = ts.xzoom
-        self.yzoom = ts.yzoom
+        d = self.__dict__
 
-        self.xaround = ts.xaround
-        self.yaround = ts.yaround
-        self.xanchoraround = ts.xanchoraround
-        self.yanchoraround = ts.yanchoraround
+        for k in all_properties:
+            d[k] = getattr(ts, k)
 
-        self.crop = ts.crop
-        self.crop_relative = ts.crop_relative
-        self.corner1 = ts.corner1
-        self.corner2 = ts.corner2
-        self.size = ts.size
+        self.last_angle = ts.last_angle
 
-        self.xpan = ts.xpan
-        self.ypan = ts.ypan
-        self.xtile = ts.xtile
-        self.ytile = ts.ytile
-
-        self.debug = ts.debug
-        self.events = ts.events
+        # Set the position and anchor to None, so inheritance works.
+        if self.perspective is None:
+            self.xpos = None
+            self.ypos = None
+            self.xanchor = None
+            self.yanchor = None
 
         # Take the computed position properties, not the
         # raw ones.
@@ -211,68 +151,41 @@ class TransformState(renpy.object.Object):
 
         rv = { }
 
-        def diff2(prop, new, old):
+        for prop in diff2_properties:
+            new = getattr(newts, prop)
+            old = getattr(self, prop)
+
             if new != old:
                 rv[prop] = (old, new)
 
-        def diff4(prop, new, inherited_new, old, inherited_old):
+        for prop in diff4_properties:
+
+            new = getattr(newts, prop)
+            old = getattr(self, prop)
+
             if new is None:
-                new_value = inherited_new
-            else:
-                new_value = new
-
+                new = getattr(newts, "inherited_" + prop)
             if old is None:
-                old_value = inherited_old
-            else:
-                old_value = old
+                old = getattr(self, "inherited_" + prop)
 
-            if new_value != old_value:
-                rv[prop] = (old_value, new_value)
-
-        diff2("nearest", newts.nearest, self.nearest)
-        diff2("alpha", newts.alpha, self.alpha)
-        diff2("additive", newts.additive, self.additive)
-        diff2("rotate", newts.rotate, self.rotate)
-        diff2("rotate_pad", newts.rotate_pad, self.rotate_pad)
-        diff2("transform_anchor", newts.transform_anchor, self.transform_anchor)
-        diff2("zoom", newts.zoom, self.zoom)
-        diff2("xzoom", newts.xzoom, self.xzoom)
-        diff2("yzoom", newts.yzoom, self.yzoom)
-
-        diff2("xaround", newts.xaround, self.xaround)
-        diff2("yaround", newts.yaround, self.yaround)
-        diff2("xanchoraround", newts.xanchoraround, self.xanchoraround)
-        diff2("yanchoraround", newts.yanchoraround, self.yanchoraround)
-
-        diff2("subpixel", newts.subpixel, self.subpixel)
-
-        diff2("crop", newts.crop, self.crop)
-        diff2("crop_relative", newts.crop_relative, self.crop_relative)
-        diff2("corner1", newts.corner1, self.corner1)
-        diff2("corner2", newts.corner2, self.corner2)
-        diff2("size", newts.size, self.size)
-
-        diff4("xpos", newts.xpos, newts.inherited_xpos, self.xpos, self.inherited_xpos)
-        diff4("xanchor", newts.xanchor, newts.inherited_xanchor, self.xanchor, self.inherited_xanchor)
-        diff2("xoffset", newts.xoffset, self.xoffset)
-
-        diff4("ypos", newts.ypos, newts.inherited_ypos, self.ypos, self.inherited_ypos)
-        diff4("yanchor", newts.yanchor, newts.inherited_yanchor, self.yanchor, self.inherited_yanchor)
-        diff2("yoffset", newts.yoffset, self.yoffset)
-
-        diff2("xpan", newts.xpan, self.xpan)
-        diff2("ypan", newts.ypan, self.ypan)
-
-        diff2("xtile", newts.xtile, self.xtile)
-        diff2("ytile", newts.ytile, self.ytile)
-
-        diff2("debug", newts.debug, self.debug)
-        diff2("events", newts.events, self.events)
+            if new != old:
+                rv[prop] = (old, new)
 
         return rv
 
     def get_placement(self, cxoffset=0, cyoffset=0):
 
+        if self.perspective is not None:
+            return (
+                0,
+                0,
+                0,
+                0,
+                cxoffset,
+                cyoffset,
+                False,
+                )
+
         return (
             first_not_none(self.xpos, self.inherited_xpos),
             first_not_none(self.ypos, self.inherited_ypos),
@@ -384,6 +297,16 @@ class TransformState(renpy.object.Object):
 
     offset = property(get_offset, set_offset)
 
+    def get_size(self):
+        return self.xsize, self.ysize
+
+    def set_size(self, value):
+        if value is None:
+            value = (None, None)
+        self.xsize, self.ysize = value
+
+    size = property(get_size, set_size)
+
     def set_xcenter(self, value):
         self.xpos = value
         self.xanchor = 0.5
@@ -401,6 +324,7 @@ class TransformState(renpy.object.Object):
     xcenter = property(get_xcenter, set_xcenter)
     ycenter = property(get_ycenter, set_ycenter)
 
+
 class Proxy(object):
     """
     This class proxies a field from the transform to its state.
@@ -415,6 +339,7 @@ class Proxy(object):
     def __set__(self, instance, value):
         return setattr(instance.state, self.name, value)
 
+
 class Transform(Container):
     """
     Documented in sphinx, because we can't scan this object.
@@ -423,64 +348,6 @@ class Transform(Container):
     __version__ = 5
     transform_event_responder = True
 
-    # Proxying things over to our state.
-    nearest = Proxy("nearest")
-    alpha = Proxy("alpha")
-    additive = Proxy("additive")
-    rotate = Proxy("rotate")
-    rotate_pad = Proxy("rotate_pad")
-    transform_anchor = Proxy("transform_anchor")
-    zoom = Proxy("zoom")
-    xzoom = Proxy("xzoom")
-    yzoom = Proxy("yzoom")
-
-    xpos = Proxy("xpos")
-    ypos = Proxy("ypos")
-    xanchor = Proxy("xanchor")
-    yanchor = Proxy("yanchor")
-
-    xalign = Proxy("xalign")
-    yalign = Proxy("yalign")
-
-    around = Proxy("around")
-    alignaround = Proxy("alignaround")
-    angle = Proxy("angle")
-    radius = Proxy("radius")
-
-    xaround = Proxy("xaround")
-    yaround = Proxy("yaround")
-    xanchoraround = Proxy("xanchoraround")
-    yanchoraround = Proxy("yanchoraround")
-
-    pos = Proxy("pos")
-    anchor = Proxy("anchor")
-    align = Proxy("align")
-
-    crop = Proxy("crop")
-    crop_relative = Proxy("crop_relative")
-    corner1 = Proxy("corner1")
-    corner2 = Proxy("corner2")
-    size = Proxy("size")
-
-    delay = Proxy("delay")
-
-    xoffset = Proxy("xoffset")
-    yoffset = Proxy("yoffset")
-    offset = Proxy("offset")
-
-    subpixel = Proxy("subpixel")
-
-    xcenter = Proxy("xcenter")
-    ycenter = Proxy("ycenter")
-
-    xpan = Proxy("xpan")
-    ypan = Proxy("ypan")
-    xtile = Proxy("xtile")
-    ytile = Proxy("ytile")
-
-    debug = Proxy("debug")
-    events = Proxy("events")
-
     def after_upgrade(self, version):
 
         if version < 1:
@@ -517,16 +384,16 @@ class Transform(Container):
             self.replaced_response = True
 
     DEFAULT_ARGUMENTS = {
-            "selected_activate" : { },
-            "selected_hover" : { },
-            "selected_idle" : { },
-            "selected_insensitive" : { },
-            "activate" : { },
-            "hover" : { },
-            "idle" : { },
-            "insensitive" : { },
-            "" : { },
-            }
+        "selected_activate" : { },
+        "selected_hover" : { },
+        "selected_idle" : { },
+        "selected_insensitive" : { },
+        "activate" : { },
+        "hover" : { },
+        "idle" : { },
+        "insensitive" : { },
+        "" : { },
+        }
 
     # Compatibility with old versions of the class.
     active = False
@@ -539,11 +406,10 @@ class Transform(Container):
     def __init__(self,
                  child=None,
                  function=None,
-
-                 style='transform',
+                 style="default",
                  focus=None,
                  default=False,
-                 _args = None,
+                 _args=None,
 
                  **kwargs):
 
@@ -566,7 +432,7 @@ class Transform(Container):
             self.arguments = { }
 
             # Fill self.arguments with a
-            for k, v in kwargs.iteritems():
+            for k, v in kwargs.items():
 
                 prefix = ""
                 prop = k
@@ -592,7 +458,7 @@ class Transform(Container):
                         prefix = new_prefix
 
             if "" in self.arguments:
-                for k, v in self.arguments[""].iteritems():
+                for k, v in self.arguments[""].items():
                     setattr(self.state, k, v)
 
         else:
@@ -651,7 +517,7 @@ class Transform(Container):
             if d is None:
                 continue
 
-            for k, v in d.iteritems():
+            for k, v in d.items():
                 setattr(state, k, v)
 
         return None
@@ -659,10 +525,10 @@ class Transform(Container):
     def set_transform_event(self, event):
         if self.child is not None:
             self.child.set_transform_event(event)
+            self.last_child_transform_event = event
 
         super(Transform, self).set_transform_event(event)
 
-
     def take_state(self, t):
         """
         Takes the transformation state from object t into this object.
@@ -676,10 +542,16 @@ class Transform(Container):
 
         self.state.take_state(t.state)
 
+        if isinstance(self.child, Transform) and isinstance(t.child, Transform):
+            self.child.take_state(t.child)
+
+        if (self.child is None) and (t.child is not None):
+            self.add(t.child)
+            self.child_st_base = t.child_st_base
+
         # The arguments will be applied when the default function is
         # called.
 
-
     def take_execution_state(self, t):
         """
         Takes the execution state from object t into this object. This is
@@ -705,7 +577,6 @@ class Transform(Container):
         if isinstance(self.child, Transform) and isinstance(t.child, Transform):
             self.child.take_execution_state(t.child)
 
-
     def copy(self):
         """
         Makes a copy of this transform.
@@ -728,10 +599,20 @@ class Transform(Container):
 
         return rv
 
-    def _hide(self, st, at, kind):
+    def _handles_event(self, event):
 
-        if not self.child:
-            return None
+        if (event == "replaced") and (not self.active):
+            return True
+
+        if self.function is not None:
+            return True
+
+        if self.child and self.child._handles_event(event):
+            return True
+
+        return False
+
+    def _hide(self, st, at, kind):
 
         # Prevent time from ticking backwards, as can happen if we replace a
         # transform but keep its state.
@@ -743,6 +624,12 @@ class Transform(Container):
         self.st = st = st + self.st_offset
         self.at = at = at + self.at_offset
 
+        if not self.active:
+            self.update_state()
+
+        if not self.child:
+            return None
+
         if not (self.hide_request or self.replaced_request):
             d = self.copy()
         else:
@@ -751,6 +638,9 @@ class Transform(Container):
         d.st_offset = self.st_offset
         d.at_offset = self.at_offset
 
+        if not (self.hide_request or self.replaced_request):
+            d.atl_st_offset = None
+
         if kind == "hide":
             d.hide_request = True
         else:
@@ -761,6 +651,8 @@ class Transform(Container):
 
         if d.function is not None:
             d.function(d, st + d.st_offset, at + d.at_offset)
+        elif isinstance(d, ATLTransform):
+            d.execute(d, st + d.st_offset, at + d.at_offset)
 
         new_child = d.child._hide(st, at, kind)
 
@@ -775,15 +667,20 @@ class Transform(Container):
 
         return None
 
-    def set_child(self, child):
+    def set_child(self, child, duplicate=True):
 
         child = renpy.easy.displayable(child)
 
-        if child._duplicatable:
+        if duplicate and child._duplicatable:
             child = child._duplicate(self._args)
+            child._unique()
+
+        if child._duplicatable:
+            self._duplicatable = True
 
         self.child = child
         self.children = [ child ]
+
         self.child_st_base = self.st
 
         child.per_interact()
@@ -795,6 +692,8 @@ class Transform(Container):
         This updates the state to that at self.st, self.at.
         """
 
+        # NOTE: This function is duplicated (more or less) in ATLTransform.
+
         self.hide_response = True
         self.replaced_response = True
 
@@ -811,8 +710,9 @@ class Transform(Container):
 
         self.active = True
 
-
     # The render method is now defined in accelerator.pyx.
+    def render(self, width, height, st, at):
+        return transform_render(self, width, height, st, at)
 
     def event(self, ev, x, y, st):
 
@@ -828,7 +728,7 @@ class Transform(Container):
         if not offsets:
             return None
 
-        for i in xrange(len(self.children)-1, -1, -1):
+        for i in range(len(self.children) - 1, -1, -1):
 
             d = children[i]
             xo, yo = offsets[i]
@@ -850,12 +750,8 @@ class Transform(Container):
         if child is None:
             child = self.child
 
-        # If we don't have a child for some reason, set it to null.
-        if child is None:
-            child = get_null()
-        else:
-            if child._duplicatable:
-                child = child._duplicate(_args)
+        if (child is not None) and (child._duplicatable):
+            child = child._duplicate(_args)
 
         rv = Transform(
             child=child,
@@ -868,6 +764,12 @@ class Transform(Container):
 
         return rv
 
+    def _unique(self):
+        if self.child and self.child._duplicatable:
+            self._duplicatable = True
+        else:
+            self._duplicatable = False
+
     def get_placement(self):
 
         if not self.active:
@@ -940,27 +842,194 @@ class Transform(Container):
     _duplicatable = True
 
     def _duplicate(self, args):
+
+        if args and args.args:
+            args.extraneous()
+
+        if not self._duplicatable:
+            return self
+
         rv = self(_args=args)
         rv.take_execution_state(self)
+        rv._unique()
+
         return rv
 
-    def _show(self):
-        self.update_state()
+    def _in_current_store(self):
+
+        if self.child is None:
+            return self
+
+        child = self.child._in_current_store()
+        if child is self.child:
+            return self
+        rv = self()
+        rv.take_execution_state(self)
+        rv.child = child
+        rv._unique()
+
+        return rv
 
-Transform.render = types.MethodType(renpy.display.accelerator.transform_render, None, Transform)
 
 class ATLTransform(renpy.atl.ATLTransformBase, Transform):
 
     def __init__(self, atl, child=None, context={}, parameters=None, **properties):
         renpy.atl.ATLTransformBase.__init__(self, atl, context, parameters)
-        Transform.__init__(self, child=child, function=self.execute, **properties)
+        Transform.__init__(self, child=child, **properties)
 
         self.raw_child = self.child
 
+    def update_state(self):
+        """
+        This updates the state to that at self.st, self.at.
+        """
+
+        self.hide_response = True
+        self.replaced_response = True
+
+        fr = self.execute(self, self.st, self.at)
+
+        # Order a redraw, if necessary.
+        if fr is not None:
+            renpy.display.render.redraw(self, fr)
+
+        self.active = True
+
     def __repr__(self):
         return "<ATL Transform {:x} {!r}>".format(id(self), self.atl.loc)
 
-    def _show(self):
-        super(ATLTransform, self)._show()
-        self.execute(self, self.st, self.at)
 
+# Names of transform properties, and if the property should be handles with
+# diff2 or diff2.
+all_properties = set()
+diff2_properties = set()
+diff4_properties = set()
+
+# Uniforms and GL properties.
+uniforms = set()
+gl_properties = set()
+
+
+def add_property(name, atl=any_object, default=None, diff=2):
+    """
+    Adds an ATL property.
+    """
+
+    if name in all_properties:
+        return
+
+    all_properties.add(name)
+    setattr(TransformState, name, default)
+    setattr(Transform, name, Proxy(name))
+    renpy.atl.PROPERTIES[name] = atl
+
+    if diff == 2:
+        diff2_properties.add(name)
+    elif diff == 4:
+        diff4_properties.add(name)
+
+
+def add_uniform(name):
+    """
+    Adds a uniform with `name` to Transform and ATL.
+    """
+
+    if not name.startswith("u_"):
+        return
+
+    if name in renpy.gl2.gl2draw.standard_uniforms:
+        return
+
+    add_property(name, diff=2)
+
+    uniforms.add(name)
+
+
+def add_gl_property(name):
+    """
+    Adds a GL property with `name` to Transform and ATL.
+    """
+
+    add_property(name, diff=None)
+
+    gl_properties.add(name)
+
+
+add_property("additive", float, 0.0)
+add_property("alpha", float, 1.0)
+add_property("blend", any_object, None)
+add_property("blur", float_or_none, None)
+add_property("corner1", (float, float), None)
+add_property("corner2", (float, float), None)
+add_property("crop", (float, float, float, float), None)
+add_property("crop_relative", bool, False)
+add_property("debug", any_object, None)
+add_property("delay", float, 0)
+add_property("events", bool, True)
+add_property("fit", str, None)
+add_property("matrixanchor", (position, position), None)
+add_property("matrixcolor", matrix, None)
+add_property("matrixtransform", matrix, None)
+add_property("maxsize", (int, int), None)
+add_property("mesh", mesh, False, diff=None)
+add_property("mesh_pad", any_object, None)
+add_property("nearest", bool_or_none, None)
+add_property("perspective", any_object, None)
+add_property("rotate", float, None)
+add_property("rotate_pad", bool, True)
+add_property("shader", any_object, None, diff=None)
+add_property("subpixel", bool, False)
+add_property("transform_anchor", bool, False)
+add_property("zoom", float, 1.0)
+
+add_property("xanchoraround", float, 0.0)
+add_property("xanchor", position, None, diff=4)
+add_property("xaround", position, 0.0)
+add_property("xoffset", float, 0.0)
+add_property("xpan", float_or_none, None)
+add_property("xpos", position, None, diff=4)
+add_property("xsize", int, None)
+add_property("xtile", int, 1)
+
+add_property("xzoom", float, 1.0)
+add_property("yanchoraround", float, 0.0)
+add_property("yanchor", position, None, diff=4)
+add_property("yaround", position, 0.0)
+add_property("yoffset", float, 0.0)
+add_property("ypan", float_or_none, None)
+add_property("ypos", position, None, diff=4)
+add_property("ysize", int, None)
+add_property("ytile", int, 1)
+add_property("yzoom", float, 1.0)
+
+add_property("zpos", float, 0.0)
+add_property("zzoom", bool, False)
+
+add_gl_property("gl_color_mask")
+add_gl_property("gl_pixel_perfect")
+add_gl_property("gl_blend_func")
+add_gl_property("gl_depth")
+add_gl_property("gl_anisostropic")
+add_gl_property("gl_mipmap")
+add_gl_property("gl_texture_wrap")
+
+ALIASES = {
+    "alignaround" : (float, float),
+    "align" : (float, float),
+    "anchor" : (position, position),
+    "angle" : float,
+    "around" : (position, position),
+    "offset" : (int, int),
+    "pos" : (position, position),
+    "radius" : float,
+    "size" : (int, int),
+    "xalign" : float,
+    "xcenter" : position,
+    "yalign" : float,
+    "ycenter" : position,
+    }
+
+renpy.atl.PROPERTIES.update(ALIASES)
+
+for name in ALIASES:
+    setattr(Transform, name, Proxy(name))
