@@ -1,4 +1,4 @@
-# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -117,7 +117,7 @@ def default_tts_function(s):
     elif renpy.windows:
 
         if renpy.config.tts_voice is None:
-            voice = "default voice"  # something that is unlikely to match.
+            voice = "default voice" # something that is unlikely to match.
         else:
             voice = renpy.config.tts_voice
 
@@ -125,6 +125,14 @@ def default_tts_function(s):
         s = s.replace('"', "")
         process = subprocess.Popen([ "wscript", fsencode(say_vbs), fsencode(s), fsencode(voice) ])
 
+    elif renpy.emscripten and renpy.config.webaudio:
+
+        try:
+            from renpy.audio.webaudio import call
+            call("tts", s)
+        except:
+            pass
+
 
 def tts(s):
     """
