@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,13 +21,11 @@
 
 # This file contains functions that load and save the game state.
 
-from __future__ import print_function
-
-import pickle
-import cPickle
-
-from cStringIO import StringIO
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+from future.utils import reraise
 
+import io
 import zipfile
 import re
 import threading
@@ -37,7 +35,9 @@ import os
 import sys
 
 import renpy
-from renpy import six
+
+import pickle
+import renpy.compat.pickle as cPickle
 
 from json import dumps as json_dumps
 
@@ -86,10 +86,10 @@ def save_dump(roots, log):
             f.write("{0: 7d} {1} = alias {2}\n".format(0, path, o_repr_cache[ido]))
             return 0
 
-        if isinstance(o, (int, float, types.NoneType, types.ModuleType, types.ClassType)):
+        if isinstance(o, (int, float, type(None), types.ModuleType, type)):
             o_repr = repr(o)
 
-        elif isinstance(o, (str, unicode)):
+        elif isinstance(o, basestring):
             if len(o) <= 80:
                 o_repr = repr(o).encode("utf-8")
             else:
@@ -102,7 +102,7 @@ def save_dump(roots, log):
             o_repr = "<" + o.__class__.__name__ + ">"
 
         elif isinstance(o, types.MethodType):
-            o_repr = "<method {0}.{1}>".format(o.im_class.__name__, o.im_func.__name__)
+            o_repr = "<method {0}.{1}>".format(o.__self__.__class__.__name__, o.__func__.__name__)
 
         elif isinstance(o, object):
             o_repr = "<{0}>".format(type(o).__name__)
@@ -112,11 +112,11 @@ def save_dump(roots, log):
 
         o_repr_cache[ido] = o_repr
 
-        if isinstance(o, (int, float, types.NoneType, types.ModuleType, types.ClassType)):
+        if isinstance(o, (int, float, type(None), types.ModuleType, type)):
             size = 1
 
-        elif isinstance(o, (str, unicode)):
-            size = len(o) / 40 + 1
+        elif isinstance(o, bytes):
+            size = len(o) // 40 + 1
 
         elif isinstance(o, (tuple, list)):
             size = 1
@@ -126,12 +126,12 @@ def save_dump(roots, log):
 
         elif isinstance(o, dict):
             size = 2
-            for k, v in o.iteritems():
+            for k, v in o.items():
                 size += 2
                 size += visit(v, "{0}[{1!r}]".format(path, k))
 
         elif isinstance(o, types.MethodType):
-            size = 1 + visit(o.im_self, path + ".im_self")
+            size = 1 + visit(o.__self__, path + ".im_self")
 
         else:
 
@@ -155,7 +155,7 @@ def save_dump(roots, log):
 
             state = get(2, { })
             if isinstance(state, dict):
-                for k, v in state.iteritems():
+                for k, v in state.items():
                     size += 2
                     size += visit(v, path + "." + k)
             else:
@@ -181,10 +181,9 @@ def save_dump(roots, log):
 
     f, _ = renpy.error.open_error_file("save_dump.txt", "w")
 
-    visit(roots, "roots")
-    visit(log, "log")
-
-    f.close()
+    with f:
+        visit(roots, "roots")
+        visit(log, "log")
 
 
 def find_bad_reduction(roots, log):
@@ -202,7 +201,7 @@ def find_bad_reduction(roots, log):
 
         seen.add(ido)
 
-        if isinstance(o, (int, float, types.NoneType, types.ClassType)):
+        if isinstance(o, (int, float, type(None), type)):
             return
 
         if isinstance(o, (tuple, list)):
@@ -212,13 +211,13 @@ def find_bad_reduction(roots, log):
                     return rv
 
         elif isinstance(o, dict):
-            for k, v in o.iteritems():
+            for k, v in o.items():
                 rv = visit(v, "{0}[{1!r}]".format(path, k))
                 if rv is not None:
                     return rv
 
         elif isinstance(o, types.MethodType):
-            return visit(o.im_self, path + ".im_self")
+            return visit(o.__self__, path + ".im_self")
 
         elif isinstance(o, types.ModuleType):
 
@@ -250,7 +249,7 @@ def find_bad_reduction(roots, log):
 
             state = get(2, { })
             if isinstance(state, dict):
-                for k, v in state.iteritems():
+                for k, v in state.items():
                     rv = visit(v, path + "." + k)
                     if rv is not None:
                         return rv
@@ -284,7 +283,6 @@ def find_bad_reduction(roots, log):
 
     return visit(log, "renpy.game.log")
 
-
 ################################################################################
 # Saving
 ################################################################################
@@ -350,24 +348,22 @@ class SaveRecord(object):
             safe_rename(filename_new, filename)
             return
 
-        zf = zipfile.ZipFile(filename_new, "w", zipfile.ZIP_DEFLATED)
+        with zipfile.ZipFile(filename_new, "w", zipfile.ZIP_DEFLATED) as zf:
+            # Screenshot.
+            if self.screenshot is not None:
+                zf.writestr("screenshot.png", self.screenshot)
 
-        # Screenshot.
-        zf.writestr("screenshot.png", self.screenshot)
+            # Extra info.
+            zf.writestr("extra_info", self.extra_info.encode("utf-8"))
 
-        # Extra info.
-        zf.writestr("extra_info", self.extra_info.encode("utf-8"))
+            # Json
+            zf.writestr("json", self.json)
 
-        # Json
-        zf.writestr("json", self.json)
+            # Version.
+            zf.writestr("renpy_version", renpy.version)
 
-        # Version.
-        zf.writestr("renpy_version", renpy.version)
-
-        # The actual game.
-        zf.writestr("log", self.log)
-
-        zf.close()
+            # The actual game.
+            zf.writestr("log", self.log)
 
         safe_rename(filename_new, filename)
 
@@ -400,7 +396,7 @@ def save(slotname, extra_info='', mutate_flag=False):
     if renpy.config.save_dump:
         save_dump(roots, renpy.game.log)
 
-    logf = StringIO()
+    logf = io.BytesIO()
     try:
         dump((roots, renpy.game.log), logf)
     except:
@@ -408,18 +404,18 @@ def save(slotname, extra_info='', mutate_flag=False):
         t, e, tb = sys.exc_info()
 
         if mutate_flag:
-            six.reraise(t, e, tb)
+            reraise(t, e, tb)
 
         try:
             bad = find_bad_reduction(roots, renpy.game.log)
         except:
-            six.reraise(t, e, tb)
+            reraise(t, e, tb)
 
         if bad is None:
-            six.reraise(t, e, tb)
+            reraise(t, e, tb)
 
-        e.args = ( e.args[0] + ' (perhaps {})'.format(bad), ) + e.args[1:]
-        six.reraise(t, e, tb)
+        e.args = (e.args[0] + ' (perhaps {})'.format(bad),) + e.args[1:]
+        reraise(t, e, tb)
 
     if mutate_flag and renpy.python.mutate_flag:
         raise SaveAbort()
@@ -440,6 +436,9 @@ def save(slotname, extra_info='', mutate_flag=False):
     clear_slot(slotname)
 
 
+# The thread used for autosave.
+autosave_thread = None
+
 # Flag that lets us know if an autosave is in progress.
 autosave_not_running = threading.Event()
 autosave_not_running.set()
@@ -448,7 +447,7 @@ autosave_not_running.set()
 autosave_counter = 0
 
 
-def autosave_thread(take_screenshot):
+def autosave_thread_function(take_screenshot):
 
     global autosave_counter
 
@@ -486,7 +485,7 @@ def autosave():
         return
 
     # That is, autosave is running.
-    if not autosave_not_running.isSet():
+    if not autosave_not_running.is_set():
         return
 
     if renpy.config.skipping:
@@ -503,6 +502,9 @@ def autosave():
     if renpy.store.main_menu:
         return
 
+    if not renpy.store._autosave:
+        return
+
     force_autosave(True)
 
 
@@ -521,13 +523,20 @@ def force_autosave(take_screenshot=False, block=False):
         If True, blocks until the autosave completes.
     """
 
+    global autosave_thread
+
     if renpy.game.after_rollback or renpy.exports.in_rollback():
         return
 
     # That is, autosave is running.
-    if not autosave_not_running.isSet():
+    if not autosave_not_running.is_set():
         return
 
+    # Join the autosave thread to clear resources.
+    if autosave_thread is not None:
+        autosave_thread.join()
+        autosave_thread = None
+
     # Do not save if we're in the main menu.
     if renpy.store.main_menu:
         return
@@ -555,18 +564,17 @@ def force_autosave(take_screenshot=False, block=False):
     autosave_not_running.clear()
 
     if not renpy.emscripten:
-        t = threading.Thread(target=autosave_thread, args=(take_screenshot,))
-        t.daemon = True
-        t.start()
+        autosave_thread = threading.Thread(target=autosave_thread_function, args=(take_screenshot,))
+        autosave_thread.daemon = True
+        autosave_thread.start()
     else:
-        import emscripten
-        emscripten.async_call(autosave_thread, take_screenshot, -1)
-
+        autosave_thread_function(take_screenshot)
 
 ################################################################################
 # Loading and Slot Manipulation
 ################################################################################
 
+
 def scan_saved_game(slotname):
 
     c = get_cache(slotname)
@@ -877,7 +885,6 @@ class Cache(object):
         self.get_screenshot()
 
 
-
 # A map from slotname to cache object. This is used to cache savegame scan
 # data until the slot changes.
 cache = { }
