@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,10 +23,11 @@
 # contained within the script file. It also handles rolling back the
 # game state to some time in the past.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 # Import the python ast module, not ours.
-ast = __import__("ast", { })
+import ast
 
 # Import the future module itself.
 import __future__
@@ -37,11 +38,38 @@ import weakref
 import re
 import sys
 import time
-
-import renpy.six as six
+import io
+import types
+import copyreg
+import functools
 
 import renpy.audio
 
+# A set of flags that indicate dict should run in future-compatible mode.
+FUTURE_FLAGS = (__future__.CO_FUTURE_DIVISION | __future__.CO_FUTURE_WITH_STATEMENT)
+
+##############################################################################
+# Monkeypatch copy_reg to work around a change in the class that RevertableSet
+# is based on.
+
+
+def _reconstructor(cls, base, state):
+    if (cls is RevertableSet) and (base is object):
+        base = set
+        state = [ ]
+
+    if base is object:
+        obj = object.__new__(cls)
+    else:
+        obj = base.__new__(cls, state)
+        if base.__init__ != object.__init__:
+            base.__init__(obj, state)
+
+    return obj
+
+
+copyreg._reconstructor = _reconstructor
+
 ##############################################################################
 # Code that implements the store.
 
@@ -52,7 +80,10 @@ import renpy.audio
 class StoreDeleted(object):
 
     def __reduce__(self):
-        return "deleted"
+        if PY2:
+            return b"deleted"
+        else:
+            return "deleted"
 
 
 deleted = StoreDeleted()
@@ -166,7 +197,7 @@ def begin_stores():
     Calls .begin on every store dict.
     """
 
-    for sd in store_dicts.itervalues():
+    for sd in store_dicts.values():
         sd.begin()
 
 
@@ -203,25 +234,27 @@ def create_store(name):
     d = store_dicts.setdefault(name, StoreDict())
     d.reset()
 
+    pyname = pystr(name)
+
     # Set the name.
-    d["__name__"] = name
-    d["__package__"] = name
+    d["__name__"] = pyname
+    d["__package__"] = pyname
 
     # Set up the default contents of the store.
     eval("1", d)
 
-    for k, v in renpy.minstore.__dict__.iteritems():
+    for k, v in renpy.minstore.__dict__.items():
         if k not in d:
             d[k] = v
 
     # Create or reuse the corresponding module.
     if name in store_modules:
-        sys.modules[name] = store_modules[name]
+        sys.modules[pyname] = store_modules[name]
     else:
-        store_modules[name] = sys.modules[name] = StoreModule(d)
+        store_modules[name] = sys.modules[pyname] = StoreModule(d)
 
     if parent:
-        store_dicts[parent][var] = sys.modules[name]
+        store_dicts[parent][var] = sys.modules[pyname]
 
 
 class StoreBackup():
@@ -279,7 +312,7 @@ def make_clean_stores():
 
     global clean_store_backup
 
-    for _k, v in store_dicts.iteritems():
+    for _k, v in store_dicts.items():
 
         v.ever_been_changed.clear()
         v.begin()
@@ -329,18 +362,16 @@ class NoRollback(object):
 
     pass
 
-
 # parents = [ ]
 
+
 def reached(obj, reachable, wait):
     """
     @param obj: The object that was reached.
-    @param path: The path from the store via which it was reached.
 
     `reachable`
         A map from id(obj) to int. The int is 1 if the object was reached
         normally, and 0 if it was reached, but inherits from NoRollback.
-
     """
 
     if wait:
@@ -351,7 +382,7 @@ def reached(obj, reachable, wait):
     if idobj in reachable:
         return
 
-    if isinstance(obj, (NoRollback, real_file)):  # @UndefinedVariable
+    if isinstance(obj, (NoRollback, io.IOBase)): # @UndefinedVariable
         reachable[idobj] = 0
         return
 
@@ -366,7 +397,7 @@ def reached(obj, reachable, wait):
 
     try:
         # Treat as fields, indexed by strings.
-        for v in vars(obj).itervalues():
+        for v in vars(obj).values():
             reached(v, reachable, wait)
     except:
         pass
@@ -381,7 +412,7 @@ def reached(obj, reachable, wait):
 
     try:
         # Treat as dict.
-        for v in obj.itervalues():
+        for v in obj.values():
             reached(v, reachable, wait)
     except:
         pass
@@ -399,33 +430,40 @@ def reached_vars(store, reachable, wait):
     the path by which the object was reached.
     """
 
-    for v in store.itervalues():
+    for v in store.values():
         reached(v, reachable, wait)
 
     for c in renpy.game.contexts:
         reached(c.info, reachable, wait)
         reached(c.music, reachable, wait)
         for d in c.dynamic_stack:
-            for v in d.itervalues():
+            for v in d.values():
                 reached(v, reachable, wait)
 
-
 # Code that replaces literals will calls to magic constructors.
 
+
+def b(s):
+    if PY2:
+        return s.encode("utf-8")
+    else:
+        return s
+
+
 class WrapNode(ast.NodeTransformer):
 
     def visit_ClassDef(self, n):
         n = self.generic_visit(n)
 
         if not n.bases:
-            n.bases.append(ast.Name(id="object", ctx=ast.Load()))
+            n.bases.append(ast.Name(id=b("object"), ctx=ast.Load()))
 
         return n
 
     def visit_SetComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__set__",
+                id=b("__renpy__set__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -437,7 +475,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__set__",
+                id=b("__renpy__set__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -448,7 +486,7 @@ class WrapNode(ast.NodeTransformer):
     def visit_ListComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__list__",
+                id=b("__renpy__list__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -462,7 +500,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__list__",
+                id=b("__renpy__list__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -473,7 +511,7 @@ class WrapNode(ast.NodeTransformer):
     def visit_DictComp(self, n):
         return ast.Call(
             func=ast.Name(
-                id="__renpy__dict__",
+                id=b("__renpy__dict__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -485,7 +523,7 @@ class WrapNode(ast.NodeTransformer):
 
         return ast.Call(
             func=ast.Name(
-                id="__renpy__dict__",
+                id=b("__renpy__dict__"),
                 ctx=ast.Load()
                 ),
             args=[ self.generic_visit(n) ],
@@ -516,7 +554,7 @@ _execute_python_hide()
     tree.body = hide.body
 
 
-unicode_re = re.compile(ur'[\u0080-\uffff]')
+unicode_re = re.compile(r'[\u0080-\uffff]')
 
 
 def unicode_sub(m):
@@ -553,16 +591,22 @@ def escape_unicode(s):
 
 
 # Flags used by py_compile.
-old_compile_flags = ( __future__.nested_scopes.compiler_flag
+old_compile_flags = (__future__.nested_scopes.compiler_flag
                       | __future__.with_statement.compiler_flag
                       )
 
-new_compile_flags = (  old_compile_flags
-                       | __future__.absolute_import.compiler_flag
-                       | __future__.print_function.compiler_flag
-                       | __future__.unicode_literals.compiler_flag
-                       )
+new_compile_flags = (old_compile_flags
+                      | __future__.absolute_import.compiler_flag
+                      | __future__.print_function.compiler_flag
+                      | __future__.unicode_literals.compiler_flag
+                      )
 
+py3_compile_flags = (new_compile_flags |
+                      __future__.division.compiler_flag)
+
+# The set of files that should be compiled under Python 2 with Python 3
+# semantics.
+py3_files = set()
 
 # A cache for the results of py_compile.
 py_compile_cache = { }
@@ -587,6 +631,95 @@ def fix_missing_locations(node, lineno, col_offset):
         fix_missing_locations(child, lineno, col_offset)
 
 
+def quote_eval(s):
+    """
+    Quotes a string for `eval`. This is necessary when it's in certain places,
+    like as part of an argument string. We need to stick \ at the end of lines
+    that don't have it already, and that aren't in triple-quoted strings.
+    """
+
+    # No newlines! No problem.
+    if "\n" not in s:
+        return s
+
+    # Characters being added to the string.
+    rv = [ ]
+
+    # Pad out the string, so we don't have to deal with quotes at the end.
+    s += "\0\0"
+
+    len_s = len(s)
+
+    # The index into the string.
+    i = 0
+
+    # Special characters, that aren't just copied into the string.
+    special = "\0\\'\"\n"
+
+    # The string currently being processed.
+    string = None
+
+    while i < len_s:
+
+        c = s[i]
+
+        # Non-special characters.
+        if c not in special:
+            start = i
+
+            while True:
+                i += 1
+                if s[i] in special:
+                    break
+
+            rv.append(s[start:i])
+            continue
+
+        # Null.
+        if c == '\0':
+            rv.append(c)
+            i += 1
+            continue
+
+        # Any escaped character passes.
+        if c == '\\':
+            rv.append(s[i:i + 2])
+            i += 2
+            continue
+
+        # String delimiters.
+        if c in '\'"':
+
+            if ((string is None) or (len(string) == 3)) and (s[i + 1] == c) and (s[i + 2] == c):
+                delim = c + c + c
+            else:
+                delim = c
+
+            if (string is not None) and (delim == string):
+                string = None
+            elif string is None:
+                string = delim
+
+            rv.append(delim)
+            i += len(delim)
+
+            continue
+
+        # Newline.
+        if c == "\n":
+            if string is None:
+                rv.append('\\')
+
+            rv.append("\n")
+            i += 1
+            continue
+
+        raise Exception("Unknown character %r (can't happen)".format(c))
+
+    # Since the last 2 characters are \0, those characters need to be stripped.
+    return "".join(rv[:-2])
+
+
 def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=True):
     """
     Compiles the given source code using the supplied codegenerator.
@@ -624,7 +757,7 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
         lineno = source.linenumber
 
     if cache:
-        key = (lineno, filename, unicode(source), mode, renpy.script.MAGIC)
+        key = (lineno, filename, str(source), mode, renpy.script.MAGIC)
 
         rv = py_compile_cache.get(key, None)
         if rv is not None:
@@ -632,7 +765,7 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
 
         rv = old_py_compile_cache.get(key, None)
         if rv is not None:
-            old_py_compile_cache[key] = rv
+            py_compile_cache[key] = rv
             return rv
 
         bytecode = renpy.game.script.bytecode_oldcache.get(key, None)
@@ -643,9 +776,11 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
             py_compile_cache[key] = rv
             return rv
 
-    source = unicode(source)
+    source = str(source)
     source = source.replace("\r", "")
-    source = escape_unicode(source)
+
+    if mode == "eval":
+        source = quote_eval(source)
 
     try:
         line_offset = lineno - 1
@@ -655,13 +790,21 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=
         else:
             py_mode = mode
 
-        try:
-            flags = new_compile_flags
-            tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
-        except:
-            flags = old_compile_flags
+        if filename in py3_files:
+
+            flags = py3_compile_flags
             tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
 
+        else:
+
+            try:
+                flags = new_compile_flags
+                tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
+            except:
+                flags = old_compile_flags
+                source = escape_unicode(source)
+                tree = compile(source, filename, py_mode, ast.PyCF_ONLY_AST | flags, 1)
+
         tree = wrap_node.visit(tree)
 
         if mode == "hide":
@@ -707,10 +850,10 @@ def py_compile_eval_bytecode(source, **kwargs):
     code = py_compile(source, 'eval', cache=False, **kwargs)
     return marshal.dumps(code)
 
-
 # Classes that are exported in place of the normal list, dict, and
 # object.
 
+
 # This is set to True whenever a mutation occurs. The save code uses
 # this to check to see if a background-save is valid.
 mutate_flag = True
@@ -718,14 +861,15 @@ mutate_flag = True
 
 def mutator(method):
 
+    @functools.wraps(method, ("__name__", "__doc__"), ())
     def do_mutation(self, *args, **kwargs):
 
         global mutate_flag
 
-        mutated = renpy.game.log.mutated  # @UndefinedVariable
+        mutated = renpy.game.log.mutated # @UndefinedVariable
 
         if id(self) not in mutated:
-            mutated[id(self)] = ( weakref.ref(self), self._clean())
+            mutated[id(self)] = (weakref.ref(self), self._clean())
             mutate_flag = True
 
         return method(self, *args, **kwargs)
@@ -817,10 +961,10 @@ class RevertableList(list):
         list.__init__(self, *args)
 
     __delitem__ = mutator(list.__delitem__)
-    if six.PY2:
+    if PY2:
         __delslice__ = mutator(list.__delslice__)
     __setitem__ = mutator(list.__setitem__)
-    if six.PY2:
+    if PY2:
         __setslice__ = mutator(list.__setslice__)
     __iadd__ = mutator(list.__iadd__)
     __imul__ = mutator(list.__imul__)
@@ -832,7 +976,8 @@ class RevertableList(list):
     reverse = mutator(list.reverse)
     sort = mutator(list.sort)
 
-    def wrapper(method):  # E0213 @NoSelf
+    def wrapper(method): # E0213 @NoSelf
+
         def newmethod(*args, **kwargs):
             l = method(*args, **kwargs)
             return RevertableList(l)
@@ -840,7 +985,7 @@ class RevertableList(list):
         return newmethod
 
     __add__ = wrapper(list.__add__)
-    if six.PY2:
+    if PY2:
         __getslice__ = wrapper(list.__getslice__)
 
     def __getitem__(self, index):
@@ -929,17 +1074,31 @@ class RevertableDict(dict):
     setdefault = mutator(dict.setdefault)
     update = mutator(dict.update)
 
-    def list_wrapper(method):  # E0213 @NoSelf
-        def newmethod(*args, **kwargs):
-            return RevertableList(method(*args, **kwargs))  # E1102
+    if PY2:
 
-        return newmethod
+        def keys(self):
+            rv = dict.keys(self)
 
-    keys = list_wrapper(dict.keys)
-    values = list_wrapper(dict.values)
-    items = list_wrapper(dict.items)
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
 
-    del list_wrapper
+            return rv
+
+        def values(self):
+            rv = dict.values(self)
+
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
+
+            return rv
+
+        def items(self):
+            rv = dict.items(self)
+
+            if (sys._getframe(1).f_code.co_flags & FUTURE_FLAGS) != FUTURE_FLAGS:
+                rv = RevertableList(rv)
+
+            return rv
 
     def copy(self):
         rv = RevertableDict()
@@ -947,7 +1106,7 @@ class RevertableDict(dict):
         return rv
 
     def _clean(self):
-        return self.items()
+        return list(self.items())
 
     def _compress(self, clean):
         return clean
@@ -968,7 +1127,7 @@ class RevertableSet(set):
             self.update(state)
 
     def __getstate__(self):
-        rv = ({ i : True for i in self}, )
+        rv = ({ i : True for i in self},)
         return rv
 
     # Required to ensure that getstate and setstate are called.
@@ -998,7 +1157,8 @@ class RevertableSet(set):
     union_update = mutator(set.update)
     update = mutator(set.update)
 
-    def wrapper(method):  # @NoSelf
+    def wrapper(method): # @NoSelf
+
         def newmethod(*args, **kwargs):
             rv = method(*args, **kwargs)
             if isinstance(rv, (set, frozenset)):
@@ -1066,6 +1226,23 @@ class RevertableObject(object):
         self.__dict__.update(compressed)
 
 
+class AlwaysRollback(RevertableObject):
+    """
+    This is a revertible object that always participates in rollback.
+    It's used when a revertable object is created by an object that
+    doesn't participate in the rollback system.
+    """
+
+    def __new__(cls, *args, **kwargs):
+        self = super(AlwaysRollback, cls).__new__(cls)
+
+        log = renpy.game.log
+        if log is not None:
+            del log.mutated[id(self)]
+
+        return self
+
+
 class RollbackRandom(random.Random):
     """
     This is used for Random objects returned by renpy.random.Random.
@@ -1089,7 +1266,10 @@ class RollbackRandom(random.Random):
         super(RollbackRandom, self).setstate(compressed)
 
     setstate = mutator(random.Random.setstate)
-    jumpahead = mutator(random.Random.jumpahead)
+
+    if PY2:
+        jumpahead = mutator(random.Random.jumpahead)
+
     getrandbits = mutator(random.Random.getrandbits)
     seed = mutator(random.Random.seed)
     random = mutator(random.Random.random)
@@ -1276,15 +1456,15 @@ class Rollback(renpy.object.Object):
 
         # Add objects reachable from the stores. (Objects that might be
         # unreachable at the moment.)
-        for changes in self.stores.itervalues():
-            for _k, v in changes.iteritems():
+        for changes in self.stores.values():
+            for _k, v in changes.items():
                 if v is not deleted:
                     reached(v, reachable, wait)
 
         # Add in objects reachable through the context.
         reached(self.context.info, reachable, wait)
         for d in self.context.dynamic_stack:
-            for v in d.itervalues():
+            for v in d.values():
                 reached(v, reachable, wait)
 
         # Add in objects reachable through displayables.
@@ -1293,14 +1473,23 @@ class Rollback(renpy.object.Object):
         # Purge object update information for unreachable objects.
         new_objects = [ ]
 
-        for o, rb in self.objects:
-            if reachable.get(id(o), 0):
-                new_objects.append((o, rb))
-                reached(rb, reachable, wait)
-            else:
-                if renpy.config.debug:
-                    print("Removing unreachable:", o, file=renpy.log.real_stdout)
-                    pass
+        objects_changed = True
+        seen = set()
+
+        while objects_changed:
+
+            objects_changed = False
+
+            for o, rb in self.objects:
+
+                id_o = id(o)
+
+                if (id_o not in seen) and reachable.get(id_o, 0):
+                    seen.add(id_o)
+                    objects_changed = True
+
+                    new_objects.append((o, rb))
+                    reached(rb, reachable, wait)
 
         del self.objects[:]
         self.objects.extend(new_objects)
@@ -1314,22 +1503,23 @@ class Rollback(renpy.object.Object):
         """
 
         for obj, roll in reversed(self.objects):
+
             if roll is not None:
                 obj._rollback(roll)
 
-        for name, changes in self.stores.iteritems():
+        for name, changes in self.stores.items():
             store = store_dicts.get(name, None)
             if store is None:
                 continue
 
-            for name, value in changes.iteritems():
+            for name, value in changes.items():
                 if value is deleted:
                     if name in store:
                         del store[name]
                 else:
                     store[name] = value
 
-        for name, changes in self.delta_ebc.iteritems():
+        for name, changes in self.delta_ebc.items():
 
             store = store_dicts.get(name, None)
             if store is None:
@@ -1527,18 +1717,18 @@ class RollbackLog(renpy.object.Object):
 
         # Update self.current.stores with the changes from each store.
         # Also updates .ever_been_changed.
-        for name, sd in store_dicts.iteritems():
+        for name, sd in store_dicts.items():
             self.current.stores[name], self.current.delta_ebc[name] = sd.get_changes(begin)
 
         # Update the list of mutated objects and what we need to do to
         # restore them.
 
-        for _i in xrange(4):
+        for _i in range(4):
 
             del self.current.objects[:]
 
             try:
-                for _k, v in self.mutated.iteritems():
+                for _k, v in self.mutated.items():
 
                     if v is None:
                         continue
@@ -1569,7 +1759,7 @@ class RollbackLog(renpy.object.Object):
 
         rv = { }
 
-        for store_name, sd in store_dicts.iteritems():
+        for store_name, sd in store_dicts.items():
             for name in sd.ever_been_changed:
                 if name in sd:
                     rv[store_name + "." + name] = sd[name]
@@ -1666,9 +1856,9 @@ class RollbackLog(renpy.object.Object):
                 fwd_name, fwd_data = self.forward[0]
 
                 if (self.current.context.current == fwd_name
-                            and data == fwd_data
-                            and (keep_rollback or self.rolled_forward)
-                        ):
+                    and data == fwd_data
+                    and (keep_rollback or self.rolled_forward)
+                    ):
                     self.forward.pop(0)
                 else:
                     del self.forward[:]
@@ -1898,6 +2088,11 @@ class RollbackLog(renpy.object.Object):
 
         begin_stores()
 
+        if self.log:
+            self.current = self.log[-1]
+        else:
+            self.current = None
+
         # Restart the context or the top context.
         if replace_context:
 
@@ -1950,7 +2145,7 @@ class RollbackLog(renpy.object.Object):
         """
 
         # Fix up old screens.
-        renpy.display.screen.before_restart()  # @UndefinedVariable
+        renpy.display.screen.before_restart() # @UndefinedVariable
 
         # Set us up as the game log.
         renpy.game.log = self
@@ -1958,7 +2153,7 @@ class RollbackLog(renpy.object.Object):
         clean_stores()
         renpy.translation.init_translation()
 
-        for name, value in roots.iteritems():
+        for name, value in roots.items():
 
             if "." in name:
                 store_name, name = name.rsplit(".", 1)
@@ -2014,18 +2209,18 @@ class RollbackLog(renpy.object.Object):
         return self.identifier_cache.get(identifier, None)
 
 
-def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"):  # @ReservedAssignment
+def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"): # @ReservedAssignment
 
     if hide:
-        locals = { }  # @ReservedAssignment
+        locals = { } # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts[store]  # @ReservedAssignment
+        globals = store_dicts[store] # @ReservedAssignment
 
     if locals is None:
-        locals = globals  # @ReservedAssignment
+        locals = globals # @ReservedAssignment
 
-    exec bytecode in globals, locals
+    exec(bytecode, globals, locals)
 
 
 def py_exec(source, hide=False, store=None):
@@ -2034,25 +2229,25 @@ def py_exec(source, hide=False, store=None):
         store = store_dicts["store"]
 
     if hide:
-        locals = { }  # @ReservedAssignment
+        locals = { } # @ReservedAssignment
     else:
-        locals = store  # @ReservedAssignment
+        locals = store # @ReservedAssignment
 
-    exec py_compile(source, 'exec') in store, locals
+    exec(py_compile(source, 'exec'), store, locals)
 
 
-def py_eval_bytecode(bytecode, globals=None, locals=None):  # @ReservedAssignment
+def py_eval_bytecode(bytecode, globals=None, locals=None): # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts["store"]  # @ReservedAssignment
+        globals = store_dicts["store"] # @ReservedAssignment
 
     if locals is None:
-        locals = globals  # @ReservedAssignment
+        locals = globals # @ReservedAssignment
 
     return eval(bytecode, globals, locals)
 
 
-def py_eval(code, globals=None, locals=None):  # @ReservedAssignment
+def py_eval(code, globals=None, locals=None): # @ReservedAssignment
     if isinstance(code, basestring):
         code = py_compile(code, 'eval')
 
@@ -2082,7 +2277,7 @@ def raise_at_location(e, loc):
     code = compile(node, filename, 'exec')
 
     # PY3 - need to change to exec().
-    exec code in { "e" : e }
+    exec(code, { "e" : e })
 
 
 # This was used to proxy accesses to the store. Now it's kept around to deal
@@ -2090,29 +2285,32 @@ def raise_at_location(e, loc):
 class StoreProxy(object):
 
     def __getattr__(self, k):
-        return getattr(renpy.store, k)  # @UndefinedVariable
+        return getattr(renpy.store, k) # @UndefinedVariable
 
     def __setattr__(self, k, v):
-        setattr(renpy.store, k, v)  # @UndefinedVariable
+        setattr(renpy.store, k, v) # @UndefinedVariable
 
     def __delattr__(self, k):
-        delattr(renpy.store, k)  # @UndefinedVariable
+        delattr(renpy.store, k) # @UndefinedVariable
 
 
-# Code for pickling bound methods.
-def method_pickle(method):
-    name = method.im_func.__name__
+if PY2:
 
-    obj = method.im_self
+    # Code for pickling bound methods.
+    def method_pickle(method):
+        name = method.im_func.__name__
 
-    if obj is None:
-        obj = method.im_class
+        obj = method.im_self
 
-    return method_unpickle, (obj, name)
+        if obj is None:
+            obj = method.im_class
 
+        return method_unpickle, (obj, name)
 
-def method_unpickle(obj, name):
-    return getattr(obj, name)
+    def method_unpickle(obj, name):
+        return getattr(obj, name)
+
+    copyreg.pickle(types.MethodType, method_pickle)
 
 # Code for pickling modules.
 
@@ -2128,8 +2326,4 @@ def module_unpickle(name):
     return __import__(name)
 
 
-import copy_reg
-import types
-
-copy_reg.pickle(types.MethodType, method_pickle)
-copy_reg.pickle(types.ModuleType, module_pickle)
+copyreg.pickle(types.ModuleType, module_pickle)
