@@ -553,12 +553,13 @@ class Layout(object):
         # data from one Layout to another.
         self.paragraph_glyphs = [ ]
 
-        width = self.scale_int(width)
-        height = self.scale_int(height)
-
+        # The virtual width and height offered to this Layout.
         self.width = width
         self.height = height
 
+        width = self.scale_int(width)
+        height = self.scale_int(height)
+
         # Figure out outlines and other info.
         outlines, xborder, yborder, xoffset, yoffset = self.figure_outlines(style)
         self.outlines = outlines
@@ -809,7 +810,15 @@ class Layout(object):
                 continue
 
             # Create the texture.
-            surf = renpy.display.pgrender.surface((sw + o, sh + o), True)
+
+            tw = sw + o
+            th = sh + o
+
+            # If not a multiple of 32, round up.
+            tw = (tw | 0x1f) + 1 if (tw & 0x1f) else tw
+            th = (th | 0x1f) + 1 if (th & 0x1f) else th
+
+            surf = renpy.display.pgrender.surface((tw, th), True)
 
             di.surface = surf
             di.override_color = color
@@ -903,13 +912,13 @@ class Layout(object):
             if self.oversample < 1:
                 return n
 
-            return n * int(self.oversample)
+            return int(n * int(self.oversample))
 
         else:
             if n == 0:
                 return 0
 
-            rv = round(n * self.oversample)
+            rv = int(round(n * self.oversample))
 
             if n < 0 and rv > -1:
                 rv = -1
@@ -980,6 +989,11 @@ class Layout(object):
                     continue
 
                 elif type == TEXT:
+
+                    if (text_displayable.mask is not None):
+                        if text != u"\u200b":
+                            text = text_displayable.mask * len(text)
+
                     line.extend(tss[-1].subsegment(text))
                     continue
 
@@ -995,6 +1009,12 @@ class Layout(object):
                     tss.pop()
 
                     if not tss:
+
+                        tag = tag[1:]
+
+                        if not renpy.text.extras.text_tags.get(tag, True):
+                            raise Exception("{/%s} isn't valid, since the %s text tag doesn't take a close tag." % (tag, tag))
+
                         raise Exception("%r closes a text tag that isn't open." % text)
 
                 elif tag == "_start":
@@ -1478,6 +1498,7 @@ class Text(renpy.display.core.Displayable):
     locked = False
 
     language = None
+    mask = None
 
     def after_upgrade(self, version):
 
@@ -1495,7 +1516,7 @@ class Text(renpy.display.core.Displayable):
             self.end = None
             self.dirty = True
 
-    def __init__(self, text, slow=None, scope=None, substitute=None, slow_done=None, replaces=None, **properties):
+    def __init__(self, text, slow=None, scope=None, substitute=None, slow_done=None, replaces=None, mask=None, **properties):
 
         super(Text, self).__init__(**properties)
 
@@ -1521,6 +1542,9 @@ class Text(renpy.display.core.Displayable):
         # The text, after substitutions.
         self.text = None
 
+        # A mask, for passwords and such.
+        self.mask = mask
+
         # Sets the text we're showing, and performs substitutions.
         self.set_text(text, scope, substitute)
 
